// $Id: Duplex.h 709 2013-08-22 15:38:47Z ge $
/// \file Duplex.h
/// \brief contains the Duplex class
///
/// $Revision: 709 $
/// \author Gerald Weber <gweberbh@gmail.com>

#ifndef GBC_EXP_DUPLEX_H
#define GBC_EXP_DUPLEX_H "$Id: Duplex.h 709 2013-08-22 15:38:47Z ge $"

#include "JobControl.h"
#include "BasePairNeighbours.h"
#include "NucleotideSequence.h"

namespace gbc {

template<class _InternalTp=char>
/// \brief Contains a sequence (std::deque) of base pairs.
///
/// This class differs from NeighbourSequence in several ways.
/// Duplex contains a std::deque of objects of type BasePair, while
/// NeighbourSequence contains a sequence of BasePairNeighbours.
class Duplex: public std::deque<BasePair<_InternalTp> >
  {
  public:
    typedef _InternalTp internal_type;             ///< type of _InternalTp
    typedef BasePair<_InternalTp> base_pair_type;  ///< type of BasePair<_InternalTp>
    typedef base_pair_type pair_type;  ///< type of BasePair<_InternalTp>
    typedef std::deque<base_pair_type> deque_type;
    typedef unsigned long int index_type;

    typename base_pair_type::SymmetryActionsType Symmetry_action;    ///< Flags which symmetry action will be performed
    NucleotideSequence<internal_type> Main_strand;          ///< Not in use yet
    NucleotideSequence<internal_type> Complementary_strand; ///< Not in use yet


  Duplex(void): Symmetry_action(base_pair_type::do_not_simplify_symmetry) {}

  Duplex(const std::string& dup,
         const typename base_pair_type::SymmetryActionsType symmetry_action=base_pair_type::do_not_simplify_symmetry)
        :Symmetry_action(symmetry_action)
    {
    std::string::size_type slash=dup.find("/");
    if (slash != std::string::npos)
      {
      std::string mn,cp;
      mn=dup.substr(0,slash);
      cp=dup.substr(slash+1);
      NucleotideSequence<internal_type> nuc1(mn.length()),nuc2(cp.length());
      nuc1 << mn;
      nuc2 << cp;
      *this=Duplex(nuc1,nuc2,symmetry_action);
      }
    else
      {
      NucleotideSequence<internal_type> nuc(dup.length());
      nuc << dup;
      *this=nuc;
      }
    }

  Duplex(const NucleotideSequence<internal_type>& nuc1, ///< Main strand
         const NucleotideSequence<internal_type>& nuc2, ///< Complimentary strand
         const typename base_pair_type::SymmetryActionsType symmetry_action=base_pair_type::do_not_simplify_symmetry)
        :Symmetry_action(symmetry_action)
    {
    deque_type::clear();
    typename NucleotideSequence<internal_type>::const_iterator n1,n2;
    for(n1=nuc1.begin(), n2=nuc2.begin(); 
       (n1 != nuc1.end()) && (n2 != nuc2.end()); ++n1, ++n2)
      {
      BasePair<internal_type> bp(*n1,*n2,Symmetry_action);
      this->push_back(bp);
      }
    if (Symmetry_action == base_pair_type::simplify_symmetry) this->reduce_to_smallest_symmetry();
    }

/*  Duplex(const nucleotide_sequence_type& fst,const nucleotide_sequence_type& scnd)
    {
    this->first=fst;
    this->second=scnd;
    }
*/

  template<class _InternalTp2>
  inline Duplex& operator+=(BasePair<_InternalTp2> bp)
    {
    if (Symmetry_action == base_pair_type::simplify_symmetry) bp.reduce_to_smallest_symmetry();
    deque_type::push_back(bp);
    return *this;
    }

  inline bool reduce_to_smallest_symmetry(void)
    {
    bool modified=false;
    typename deque_type::iterator it;
    for(it=this->begin(); it!=this->end(); ++it)
      {
      bool m=it->reduce_to_smallest_symmetry();
      modified=modified || m;
      }
    return modified;
    }

  template<class _InternalTp2>
  inline Duplex& operator+=(const Nucleotide<_InternalTp2> nuc)
    {
    BasePair<_InternalTp2> bp(nuc,Symmetry_action);
    if (Symmetry_action == base_pair_type::simplify_symmetry) bp.reduce_to_smallest_symmetry();
    deque_type::push_back(bp);
    return *this;
    }

  template<class _InternalTp2>
  /// \brief Copies a NucleotideSequence to the Duplex
  ///
  /// Each Nucleotide is converted to a BasePair and added to the
  /// end of the Duplex.
  ///
  /// \attention The previous content of the Duplex is erased
  inline Duplex& operator=(const NucleotideSequence<_InternalTp2>& nucseq)
    {
    deque_type::clear();
    typename NucleotideSequence<_InternalTp2>::const_iterator nt;
    for(nt=nucseq.begin(); nt != nucseq.end(); ++nt)
      this->operator+=(*nt);
    return *this;
    }

  inline BasePairNeighbours<_InternalTp> operator()(const index_type bg, const index_type ed) const
    {
    BasePairNeighbours<_InternalTp> nb((*this)[bg],(*this)[ed]);
    return nb;
    }

  inline NucleotideSequence<_InternalTp> main_strand(void) const
    {
    NucleotideSequence<_InternalTp> strand(this->size());
    typename deque_type::const_iterator it;
    for(it=this->begin(); it!=this->end(); ++it)
      {
      strand.push_back(it->first);
      }
    return strand;
    }

  inline NucleotideSequence<_InternalTp> complementary_strand(void) const
    {
    NucleotideSequence<_InternalTp> strand(this->size());
    typename deque_type::const_iterator it;
    for(it=this->begin(); it!=this->end(); ++it)
      {
      strand.push_back(it->second);
      }
    return strand;
    }

   inline bool self_complementary(void) const
     {
     NucleotideSequence<_InternalTp> s(main_strand()),c(complementary_strand());
     c.reverse();
     if ((std::string)s == (std::string)c) return true;
     else                                  return false;
     }

  /// Operator which returns a concatenated string of symbols representing the duplex.
  inline operator std::string(void) const
    {
    std::string first,second,out;
    typename deque_type::const_iterator it;
    for (it=this->begin(); it != this->end(); it++)
      {
      first += it->first.symbol();
      second+= it->second.symbol();
      }
    out = first + "/" + second;
    return out;
    }

  /// Extractor for printing BasePair symbols
  inline friend std::ostream& operator<<(std::ostream &out,const Duplex &dp)
    { 
    std::copy(dp.begin(),dp.end(),std::ostream_iterator<base_pair_type>(out,"/"));
    return out;
    }

  };//ends class


};//ends namespace
#endif
