// $Id: ParameterSearch.h 753 2014-04-08 13:35:13Z ge $
/// \file ParameterSearch.h
/// \brief contains the ParameterSearch function class
///
/// $Revision: 753 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_EXP_PARAMETERSEARCH_H
#define GBC_EXP_PARAMETERSEARCH_H "$Id: ParameterSearch.h 753 2014-04-08 13:35:13Z ge $"
#include "Minimization.h"
#include "GSLMinimization.h"
#include <cmath>

namespace gbc_exp
{
template<class _EvaluateTp>
class ParameterSearch
  {
  public:
    typedef _EvaluateTp evaluate_type;
    typedef typename evaluate_type::value_type value_type;
    typedef std::valarray<value_type> vector_type;
    typedef std::map<std::string,value_type>     map_type;
    typedef std::map<size_t,map_type>        all_map_type;
    class files_class
      {
      public:
      std::ofstream Result;
      std::ofstream Parameter;
      std::ofstream Verification;
      std::ofstream Variation;
      std::ofstream Log;
      std::ofstream Regression;
      std::ofstream Evolution;
      std::string BaseName, VersionBase;
      int Version;
      bool Make_proceed_file;
      bool Make_files;
      bool Versioned;
      
      files_class(void) : Version(0), Make_proceed_file(true), Make_files(true), Versioned(false) {};
      
      inline void next_version(void)
        {
        close_all();
        Version++;
        std::stringstream ver;
        ver << "." << std::setw(5) << std::setfill('0') << Version;
        VersionBase=ver.str();
        open_all_version();
        }
      inline void overwrite(void)
        {
        close_all();
        if (Versioned) open_all_version(); else open_all();
        }
      inline void open(std::string bsname)
        {
        BaseName=bsname;
        if (Versioned) open_all_version(); else open_all();
        }
      inline void open_all(void)
        {
        if (not Make_files) return;
        Result.open((BaseName+".dat").c_str());
        Parameter.open((BaseName+".par").c_str());
        Verification.open((BaseName+".ver").c_str());
        Variation.open((BaseName+".var").c_str());
        Log.open((BaseName+".log").c_str());
        Regression.open((BaseName+".reg").c_str());
        Evolution.open((BaseName+".average_std_evolution").c_str());

        std::ofstream ProcFile;
        if (Make_proceed_file) 
          {
          ProcFile.open((BaseName+".proceed").c_str());
          ProcFile << "1" << std::endl;
          ProcFile.close();
          }
        }
      inline void open_all_version(void)
        {
        if (not Make_files) return;
        Result.open((BaseName+".dat"+VersionBase).c_str());
        Parameter.open((BaseName+".par"+VersionBase).c_str());
        Verification.open((BaseName+".ver"+VersionBase).c_str());
        Variation.open((BaseName+".var"+VersionBase).c_str());
        Log.open((BaseName+".log"+VersionBase).c_str());
        Regression.open((BaseName+".reg"+VersionBase).c_str());
        }

      inline void close_all(void)
        {
        Result.close();
        Parameter.close();
        Verification.close();
        Variation.close();
        Log.close();
        Regression.close();
        Evolution.close();
        }
      inline bool proceed(void)
        {
        std::ifstream Proceed;
        Proceed.open((BaseName+".proceed").c_str());
        bool proc;
        Proceed >> proc;
        Proceed.close();
        return proc;
        }
      };
    

    enum minimization_engine_enum {GBC,GSL};
    minimization_engine_enum MinimizationEngine;
    /// The minimization class takes as template a class which performs the merit function evaluation
    Minimization<ParameterSearch<evaluate_type> > Minimize;
    GSLMinimization<ParameterSearch<evaluate_type> > GSLMinimize;

    evaluate_type Evaluation; ///< This is the function that will run in minimization

    bool Debug;
    bool Random_search;
    bool Proceed;
    bool Avoid_negative_parameters;
    bool Randomise_initial_parameters;
    bool Randomise_dataset;
    bool Generate_files_for_each_round;
    value_type Initial_parameter_variation;
    int MaxEvaluations;
    int MaxMinimizations;
    int MaxRandomMinimizations; ///< Maximum number of random regression minimization, 0 means go on forever
    long int Evaluations;
    value_type Total_average_difference;
    value_type Total_standard_deviation_difference;
    value_type Limit_average_difference;
    value_type Limit_standard_deviation_difference;
    map_type Variable_map;
    map_type Best_variable_map;
    map_type Average_variable_map;
    map_type Previous_average_variable_map;
    map_type Standard_deviation_variable_map;
    map_type Previous_standard_deviation_variable_map;
    all_map_type Complete_variable_map;
    vector_type Best_vector;
    value_type  Best_value;
    std::string BaseFileName;
    files_class File, PermanentFile, BestValueFile;
    Range<value_type> Parameter_range;

    ParameterSearch(void)
      : MinimizationEngine(GBC), Minimize(), GSLMinimize(), Evaluation(), Debug(false), Random_search(false), Proceed(true), Avoid_negative_parameters(true),
        Randomise_initial_parameters(false), Randomise_dataset(true), Generate_files_for_each_round(true), Initial_parameter_variation(value_type()),
        MaxEvaluations(1000), MaxMinimizations(1), MaxRandomMinimizations(0), Evaluations(), 
        Total_average_difference(), Total_standard_deviation_difference(), Limit_average_difference(), Limit_standard_deviation_difference()
         {}
        
    void minimization_method(std::string option)
      {
      if (option == std::string("gbc")) MinimizationEngine=GBC;
      else 
        {
        if (option == std::string("gsl")) MinimizationEngine=GSL;
        }
      }
        
      

    /// \brief Assing list of parameters which will vary during the minimization
    ///
    /// This will assign the current definition of the parameter to its 
    /// corresponding position in Variable_map.
    /// The initial value of Variable_map is whatever was stored in ParameterMap initially.
    inline void assign(const std::deque<std::string>& var) ///< List of parameters, usually given on command line
      {
      typename std::deque<std::string>::const_iterator var_list_it;
      for(var_list_it=var.begin(); var_list_it!= var.end(); ++var_list_it)
        {
        if (*var_list_it == std::string("T")) 
          Variable_map[*var_list_it]=Evaluation.Calculation.temperature(); // If parameter is simply a T, assing temperature to it.
        else 
          Evaluation.Calculation.parameter_map.get_parameter(*var_list_it,Variable_map[*var_list_it]);
        if (Debug) std::cout << __FILE__ << " " << __LINE__ << "\n" 
                             << *var_list_it << " "<< Variable_map[*var_list_it] << "\n" << std::endl;
        }
      } 

    /// \brief Reassings the initial values of VariableMap from ParameterMap.
    inline void variable_map_reassign(void)
      {
      typename map_type::iterator it;
      for(it=Variable_map.begin(); it!= Variable_map.end(); ++it)
        {
        if (it->first == std::string("T")) 
          it->second=Evaluation.Calculation.temperature(); // If parameter is simply a T, assing temperature to it.
        else 
          Evaluation.Calculation.parameter_map.insert_parameter(it->first,it->second);
        if (Debug) std::cout << __FILE__ << " " << __LINE__ << "\n" 
                             << it->first << " "<< Variable_map[it->first] << "\n" << std::endl;
        }
      } 

    /// \brief Randomize the initial values of VariableMap from ParameterMap.
    inline void variable_map_randomise(double rnd)
      {
      typename map_type::iterator it;
      for(it=Variable_map.begin(); it!= Variable_map.end(); ++it)
        {
        if (it->first == std::string("T")) 
          it->second=Evaluation.Calculation.temperature(); // If parameter is simply a T, assing temperature to it.
        else
	  {
          double dice=linear_ran(2*rnd,0.5);
	  it->second *=(1.0+dice);
          Evaluation.Calculation.parameter_map.insert_parameter(it->first,it->second);
          if (Debug) std::cout << __FILE__ << " " << __LINE__ << "\n" 
                               << it->first << "<--->" << it->second << "\n" << std::endl;
	  }
        if (Debug) std::cout << __FILE__ << " " << __LINE__ << "\n" 
                             << it->first << " "<< Variable_map[it->first] << "\n" << std::endl;
        }
      } 

    inline void basefile(std::string bsname)
      {
      BaseFileName=bsname;
      }

    /// \brief Assigns the vector from the Variable_map
    inline void matrix2vector(vector_type& vec)
      {
      if (vec.size() != Variable_map.size()) vec.resize(Variable_map.size());
      typename map_type::iterator it;
      size_t n=0;
      for(it=Variable_map.begin(), n=0; it != Variable_map.end(); ++it, ++n)
        vec[n]=it->second;
      }

    /// \brief Assigns the contents of a vector to the Variable_map
    ///
    /// The vector is assigned in the same order as the variables are stored.
    inline void vector2var(const vector_type& vec)
      {
      typename map_type::iterator it;
      size_t n=0;
      for(it=Variable_map.begin(), n=0; it != Variable_map.end(); ++it, ++n)
        {
        it->second=vec[n];
        }
      }

    inline vector_type randomise(value_type scale,const vector_type& vec)
      {
      vector_type nvec(vec.size());
      size_t n=0;
      for(n=0; n < vec.size(); ++n)
        {
        double dice=rand()/static_cast<double>(RAND_MAX)-0.5;
        nvec[n]=vec[n]+dice*scale*vec[n];
        }
      return nvec;
      }

    inline vector_type random_search(value_type scale, const vector_type& vec)
      {
      vector_type nvec=randomise(scale,vec);
      value_type nv=evaluate(nvec);
      int evaluations=1;
      while ((nv > Best_value) && (evaluations < MaxEvaluations))
        {
        nvec=randomise(scale,vec);
        nv=evaluate(nvec);
        evaluations++;
        }
      if (nv < Best_value) 
        {
        File.Log << "random_search: " << evaluations << " ";
        print_res(File.Log,nv);
        return nvec;
        }
      else return vec;
      }


    /// \brief Evaluates the function from the vector or parameters
    ///
    /// First we assign the new parameters to the parameter map of the HeterogenousTM class.
    /// At the same time we update the Variable_map.
    /// Then the HeterogenousTM issues a get_parameters() call which load the modified
    /// parameters into the ModelPotentials. Only then we do the Evaluation which
    /// is defined in the SearchAction class.
    /// \Attention This returns Eavluate*negative_parameter_penalty
    inline value_type evaluate(const vector_type& vec)
     {
     value_type negative_parameter_penalty=1.0;
     typename map_type::iterator var_map_it;
     size_t n=0;
     for(var_map_it=Variable_map.begin(), n=0; var_map_it != Variable_map.end(); ++var_map_it, ++n)
       {
       if (var_map_it->first == std::string("T")) 
         Evaluation.Calculation.temperature(vec[n]);
       else
         Evaluation.Calculation.parameter_map.set_parameters(var_map_it->first,vec[n]);
       var_map_it->second=vec[n];
       if (Avoid_negative_parameters && (vec[n] < 0.0)) negative_parameter_penalty += 1.0;
       }
     //if (Debug) std::cout << Calculation << std::endl;
     if (Debug) std::cout << __FILE__ << " " << __LINE__ << "\n" 
                          << Evaluation.Calculation.parameter_map << std::endl;
     Evaluation.Calculation.get_parameters();
     Evaluations++;
     return negative_parameter_penalty*Evaluation.evaluate();
     }

    /// \brief Evaluates the function from the Variable_map
    ///
    /// The Variable_map is save into the ParameterMap if HeterogenousTM.
    /// Then the HeterogenousTM issues a get_parameters() call which load the modified
    /// parameters into the ModelPotentials. Only then we do the Evaluation which
    /// is defined in the SearchAction class.
    inline value_type evaluate(void)
     {
     typename map_type::const_iterator var_map_it;
     for(var_map_it=Variable_map.begin(); var_map_it != Variable_map.end(); ++var_map_it)
       {
       Evaluation.Calculation.parameter_map.set_parameters(var_map_it->first,var_map_it->second);
       }
     if (Debug) std::cout << __FILE__ << " " << __LINE__ << "\n" 
                          << std::cout << Evaluation.Calculation.parameter_map << std::endl;
     Evaluation.Calculation.get_parameters();
     return Evaluation.evaluate();
     }

    /// \brief Scans the parameters for the Variable_map 
    ///
    /// We increment the parameters in Variable_map sequentially and run evaluate.
    /// \Attention This function can be very time consuming if the ranges have
    /// a large number of steps or if there are too many variables in Variable_map.
    inline void scan_parameters(Range<value_type>& prange)
      {
      value_type Best_sqrdiff=1e+12, sqrdiff;

      File.open(BaseFileName);
      typename std::vector<Range<value_type> > ranges;
      prange.reset();
      typename map_type::iterator it;
      for(it=Variable_map.begin(); it != Variable_map.end(); ++it)
        {
        ranges.push_back(prange);
        it->second=prange.begin();
        File.Log << it->first << " " << prange << " " << prange.steps() << " ";
        }
      File.Log << std::endl;

      for(prange.reset(); !prange.finished(); ++prange) File.Log << prange.current() << " ";
      File.Log << std::endl;

      it=Variable_map.begin(); ++it; File.Log << it->second << " ";
      sqrdiff=evaluate();
      File.Log << sqrdiff << " ";
      if (sqrdiff < Best_sqrdiff)
        {
        Best_sqrdiff=sqrdiff;
        Best_variable_map=Variable_map;
        }

      typename std::vector<Range<value_type> >::iterator next=ranges.begin();
      ++next;
      typename std::vector<Range<value_type> >::iterator loop=ranges.begin();
      bool finished=false;
      typename map_type::iterator variable;
      variable=Variable_map.begin();
      ++variable;
      while (!finished)
        {
        if (loop->finished())
          {
          File.Log << std::endl;
          loop->reset(); Variable_map.begin()->second=loop->current();
          if (next != ranges.end())
            {
            if (next->finished())
              {
              ++next; ++variable; if (next == ranges.end()) break;
              }
            else
              {
              next->next_step(); variable->second=next->current();
              File.Log << variable->second << " ";
              }
            }
          else break;
          }
        else
          {
          loop->next_step(); Variable_map.begin()->second=loop->current();
          }

        if (!loop->finished() && !next->finished())
          {
          sqrdiff=evaluate();
          File.Log << sqrdiff << " ";
          if (sqrdiff < Best_sqrdiff)
            {
            Best_sqrdiff=sqrdiff;
            Best_variable_map=Variable_map;
            }
          }

        }

      Variable_map=Best_variable_map;
      evaluate();

      print_var(File.Variation,"");
      print_par(File.Parameter,"");
      print_best_value(File.Verification);
      Evaluation.print_verification(File.Verification);

      }

    /// \brief Scans the parameters for the Variable_map 
    ///
    /// We increment the parameters in Variable_map sequentially and run evaluate.
    /// \Attention This function can be very time consuming if the ranges have
    /// a large number of steps or if there are too many variables in Variable_map.
    inline void scan_parameters(Range<value_type>& prange1, Range<value_type>& prange2)
      {
      value_type Best_sqrdiff=1e+12, sqrdiff;

      File.open(BaseFileName);
      typename std::vector<Range<value_type> > ranges;
      prange1.reset();
      prange2.reset();
      typename map_type::iterator it=Variable_map.begin(); 
      ranges.push_back(prange1);
      it->second=prange1.begin();
      File.Log << it->first << " " << prange1 << " " << prange1.steps() << " ";
      ++it;
      ranges.push_back(prange2);
      it->second=prange2.begin();
      File.Log << it->first << " " << prange2 << " " << prange2.steps() << " ";
      File.Log << std::endl;

      for(prange1.reset(); !prange1.finished(); ++prange1) File.Log << prange1.current() << " ";
      File.Log << std::endl;

      it=Variable_map.begin(); ++it; File.Log << it->second << " ";
      sqrdiff=evaluate();
      File.Log << sqrdiff << " ";
      if (sqrdiff < Best_sqrdiff)
        {
        Best_sqrdiff=sqrdiff;
        Best_variable_map=Variable_map;
        }

      typename std::vector<Range<value_type> >::iterator next=ranges.begin();
      ++next;
      typename std::vector<Range<value_type> >::iterator loop=ranges.begin();
      bool finished=false;
      typename map_type::iterator variable;
      variable=Variable_map.begin();
      ++variable;
      while (!finished)
        {
        if (loop->finished())
          {
          File.Log << std::endl;
          loop->reset(); Variable_map.begin()->second=loop->current();
          if (next != ranges.end())
            {
            if (next->finished())
              {
              ++next; ++variable; if (next == ranges.end()) break;
              }
            else
              {
              next->next_step(); variable->second=next->current();
              File.Log << variable->second << " ";
              }
            }
          else break;
          }
        else
          {
          loop->next_step(); Variable_map.begin()->second=loop->current();
          }

        if (!loop->finished() && !next->finished())
          {
          sqrdiff=evaluate();
          File.Log << sqrdiff << " ";
          if (sqrdiff < Best_sqrdiff)
            {
            Best_sqrdiff=sqrdiff;
            Best_variable_map=Variable_map;
            }
          }

        }

      Variable_map=Best_variable_map;
      evaluate();

      print_var(File.Variation,"");
      print_par(File.Parameter,"");
      print_best_value(File.Verification);
      Evaluation.print_verification(File.Verification);

      }

    inline void minimize(Range<double>& clength,const double& limit)
      {
      //Evaluation.Debug=true;
      value_type cl=clength.current();
      vector_type  p; matrix2vector(p);
      Best_vector.resize(p.size());
      Best_value=1e+10;
      value_type previous=Best_value, diff=1e+10;
      int improove=0;
      evaluate(p);
      while((diff > limit) && (improove < MaxMinimizations))
        {
        switch (MinimizationEngine)
          {
          case GBC: Minimize.reset(*this,p,cl);
                    Minimize.MaxEvaluations=MaxEvaluations;
                    Minimize.run(limit);
                    if (Best_value > Minimize.best_value())
                      {            
                      Best_value=Minimize.best_value();
                      Best_vector=Minimize.best();
                      }
                    break;
          case GSL: GSLMinimize.reset(*this,p,cl);
                    GSLMinimize.MaxEvaluations=MaxEvaluations;
                    GSLMinimize.run(limit);
                    if (Best_value > GSLMinimize.best_value())
                      {            
                      Best_value=GSLMinimize.best_value();
                      Best_vector=GSLMinimize.best();
                      }
                    break;
          };
        diff=fabs(Best_value-previous);
        previous=Best_value;
        improove++;
        if (Random_search) p=random_search(0.5,Best_vector);
        else p=Best_vector;
        }
      if (diff <= limit) File.Log << "*** No difference between the last two rounds, stopping minimization altogether" << std::endl; 
      File.Log << "*** Local minimization stopped at round " << improove << " (maximum rounds was "<< MaxMinimizations << ")" << std::endl;
      evaluate(Best_vector);
      vector2var(Best_vector);
      }

   inline void minimization(std::string par_id, double salt, Range<double>& clength,const double& limit)
     {
     File.open(BaseFileName);
     print_par(File.Verification,par_id);
     File.Verification << "Calculation:\n" << Evaluation.Calculation << std::endl;
     File.Verification << "Calculation.parameter_map:\n" << Evaluation.Calculation.parameter_map << std::endl;
     Minimize.LogFile=&File.Log;
     GSLMinimize.LogFile=&File.Log;
     Evaluation.print_verification(File.Verification);
     minimize(clength,limit);
     print_res(File.Result,salt);
     print_var(File.Variation,par_id);
     print_par(File.Parameter,par_id);
     print_best_value(File.Verification);
     Evaluation.print_verification(File.Verification);
     Evaluation.print_regression(File.Regression);
     File.close_all();
     }

   /// \brief Performs a minimization on randomized datasets
   inline void minimization_random(int random_seed,        ///< Randomization seed
                                   std::string par_id,     ///< parameter identification
                                   double salt,            ///< salt concentration, only used for printing out results, has otherwise no effect
                                   Range<double>& clength, ///< minimization parameter characteristic length
                                   const double& limit)    ///< minimization parameter limit below which we stop searching for better values
     {
     ParameterMap<value_type> pm=Evaluation.Calculation.parameter_map; //Here we store the initial parameters
     map_type vm=Variable_map; //Here we store the initial Variable Map
     srand(random_seed);
     PermanentFile.open(BaseFileName);
     PermanentFile.Log << "Random seed=" << random_seed << std::endl;
     File.BaseName=BaseFileName;
     File.Versioned=true;
     File.Make_files=Generate_files_for_each_round;
     BestValueFile.BaseName=BaseFileName;
     BestValueFile.Versioned=true;
     BestValueFile.VersionBase=".best";
     Proceed=PermanentFile.proceed();
     JobControl::status("proceed",Proceed);
     Evaluation.print_verification(PermanentFile.Verification);
     PermanentFile.Verification 
       << Evaluation.Regression.dataset.Temperature.error << " "
       << Evaluation.Regression.dataset.Enthalpy.error << " "
       << Evaluation.Regression.dataset.Entropy.error << std::endl;
     value_type Global_best_value=1e+10;
     while(Proceed)
       {
       File.next_version();
       if (Randomise_initial_parameters)
         {
         //Evaluation.Calculation.parameter_map.randomise(Initial_parameter_variation);
         //variable_map_reassign();
         variable_map_randomise(Initial_parameter_variation);
         std::cout << "Initial variables randomised for minimization round " << File.Version << std::endl;
         }
       print_par(File.Verification,par_id);
       File.Verification << "Calculation:\n" << Evaluation.Calculation << std::endl;
       File.Verification << "Calculation.parameter_map:\n" << Evaluation.Calculation.parameter_map << std::endl;
       Minimize.LogFile=&File.Log;
       Minimize.AlwaysPrintLog=true;
       GSLMinimize.LogFile=&File.Log;
       GSLMinimize.AlwaysPrintLog=true;
       if (Randomise_dataset)
         {
         Evaluation.Regression.dataset.randomise_dataset();
         File.Verification << Evaluation.Regression.dataset.Temperature.std_deviation << " "
                           << Evaluation.Regression.dataset.Enthalpy.std_deviation << " "
                           << Evaluation.Regression.dataset.Entropy.std_deviation << std::endl;
         PermanentFile.Verification << Evaluation.Regression.dataset.Temperature.std_deviation << " "
                           << Evaluation.Regression.dataset.Enthalpy.std_deviation << " "
                           << Evaluation.Regression.dataset.Entropy.std_deviation << std::endl;
	 }
       Evaluation.print_verification(File.Verification);
       minimize(clength,limit); //here is the part that takes time to run
       
       Complete_variable_map[File.Version]=Variable_map;

       print_res(File.Result,salt);
       print_res(PermanentFile.Result,salt);
       print_var(File.Variation,par_id);
       print_par(File.Parameter,par_id);
       print_best_value(File.Verification);
       Evaluation.print_verification(File.Verification);
       Evaluation.print_regression(File.Regression);
       
       if (Global_best_value > Best_value)
         {
         std::cout << "New best set of parameters at round " << File.Version << " with chi2 = " << Best_value << std::endl;
         Global_best_value = Best_value;
         BestValueFile.overwrite();
         print_res(BestValueFile.Result,salt);
         print_var(BestValueFile.Variation,par_id);
         print_par(BestValueFile.Parameter,par_id);
         print_best_value(BestValueFile.Verification);
         Evaluation.print_verification(BestValueFile.Verification);
         Evaluation.print_regression(BestValueFile.Regression);
         BestValueFile.close_all();
         }
       
       //From here we reset everything to start over
       Evaluation.Calculation.parameter_map=pm; // restore initial parameters
       Evaluation.Calculation.get_parameters();
       Variable_map=vm; //restore initial variable map
       JobControl::summary(PermanentFile.Log) << std::endl;
       Proceed=PermanentFile.proceed();
       JobControl::status("proceed",Proceed);
       this->save_av_par_file(par_id,File.Version);
       
       if (MaxRandomMinimizations and (File.Version >= MaxRandomMinimizations)) 
         {
         std::cout << "Maximum number of randomized local minimizations reached (max=" << File.Version << ")" << std::endl;
         break;
         }
       if ((Total_average_difference <  Limit_average_difference) and (File.Version >= 10)) 
         {
         std::cout << "Limit for total average difference reached " << Total_average_difference
         << " limit= " << Limit_average_difference << ", stoped at round " << File.Version << std::endl;
         break;
         }
      if ((Total_standard_deviation_difference <  Limit_standard_deviation_difference) and (File.Version >= 10)) 
         {
         std::cout << "Limit for total standard deviation difference reached " << Total_standard_deviation_difference
         << " limit= " << Limit_standard_deviation_difference << ", stoped at round " << File.Version << std::endl;
         break;
         }
         
       }
     }
     
   inline void make_averages(void)
     {
     typename all_map_type::iterator mp;
     typename map_type::iterator var;
     Average_variable_map.clear();
     Standard_deviation_variable_map.clear();
     for(mp = Complete_variable_map.begin(); mp != Complete_variable_map.end(); ++mp)
       {
       for(var=mp->second.begin(); var != mp->second.end(); ++var)
         {
         Average_variable_map[var->first] += var->second;
         }
       }
    for(var=Average_variable_map.begin(); var != Average_variable_map.end(); ++var)
      {
      var->second /= Complete_variable_map.size();
      }
    for(mp = Complete_variable_map.begin(); mp != Complete_variable_map.end(); ++mp)
       {
       for(var=mp->second.begin(); var != mp->second.end(); ++var)
         {
         Standard_deviation_variable_map[var->first] += pow(Average_variable_map[var->first]-var->second,2);
         }
       }
    for(var=Standard_deviation_variable_map.begin(); var != Standard_deviation_variable_map.end(); ++var)
      {
      var->second = sqrt(var->second/Complete_variable_map.size());  
      }
    }

   inline void print_var(std::ostream& out, std::string ident)
      {
      out << ident << std::endl;
      typename map_type::iterator it;
      for(it=Variable_map.begin(); it != Variable_map.end(); ++it)
        out << it->first << " " << it->second << std::endl;
      }

    inline void print_par(std::ostream& out, std::string ident)
      {
      out << ident << std::endl;
      out << Evaluation.Calculation.parameter_map << std::endl;
      }

    inline void save_av_par_file(std::string indent, int version)
      {
      make_averages();
      std::ofstream out;
      out.open((BaseFileName+".par.average").c_str());
      out << indent << "-average-" << version << std::endl;
      typename map_type::iterator var;
      for(var=Average_variable_map.begin(); var != Average_variable_map.end(); ++var)
        {
        out << var->first << " " << var->second << " +/- " << Standard_deviation_variable_map[var->first];
        if (Previous_average_variable_map[var->first] != 0)
          {
          value_type average_difference=fabs((Previous_average_variable_map[var->first]-var->second)/Previous_average_variable_map[var->first]);
          out << " [" << average_difference;
          Total_average_difference += average_difference;
          }
        if (Previous_standard_deviation_variable_map[var->first] != 0)
          {
          value_type std_difference=fabs((Previous_standard_deviation_variable_map[var->first]-Standard_deviation_variable_map[var->first])/Previous_standard_deviation_variable_map[var->first]);
          out << " " << std_difference << "]";
          Total_standard_deviation_difference += std_difference;
          }
        out << std::endl;
        
        }
      out.close();
      
      Previous_average_variable_map=Average_variable_map;
      Previous_standard_deviation_variable_map=Standard_deviation_variable_map;
     
      Total_average_difference /= Average_variable_map.size();
      Total_standard_deviation_difference /= Average_variable_map.size();
      
      PermanentFile.Evolution << Complete_variable_map.size() << " " << Total_average_difference << " " << Total_standard_deviation_difference << std::endl;
      }

    inline void print_head(std::ostream& out)
      {
      typename map_type::iterator it;
      out << Best_value << "Best ";
      for(it=Variable_map.begin(); it != Variable_map.end(); ++it)
        out << it->first << " ";
      out << std::endl;
      }


    inline void print_res(std::ostream& out, value_type sl)
      {
      typename map_type::iterator it;
      if (sl != value_type()) out << sl << " ";
      out << Best_value << " ";
      for(it=Variable_map.begin(); it != Variable_map.end(); ++it)
        out << it->second << " ";
      out << std::endl;
      }

    inline void print_best_value(std::ostream& out)
      {
      out << Best_value << std::endl;
      }

  };
}
#endif
