// $Id: var-gibbs.cpp 728 2013-11-06 15:58:33Z ge $
/// \file var-gibbs.cpp
/// \brief program for searching parameters of Gibbs Free Energy Models
/// \author Gerald Weber <gweberbh@gmail.com>
/// $Revision: 728 $

#include "JobControl.h"
#include "Options.h"
#include "NeighbourSequence.h"
#include <unistd.h>
#include "FreeEnergyModel.h"
#include "ParameterSearch.h" 
#include "GibbsMinimization.h"

using namespace gbc;
using namespace std;
namespace gbc_test=gbc_exp;
using namespace gbc_test;

typedef double value_type;

typedef ParameterSearch<GibbsMinimization<IndividualNearestNeighbours<double> > > min_function_type;
 

// We need to declare an object of type min_function_type to avoid compiler errors
// of inclomplete type
min_function_type place_holder;
template<> min_function_type *GSLMinimization<min_function_type>::Function=NULL;


int main(int argc,char *argv[])
  {
  JobControl::status("running",true);
  JobControl::write_log("$Id: var-gibbs.cpp 728 2013-11-06 15:58:33Z ge $");
 
  Option<string>         output_name("o","output files basename, resulting files will be appended with .dat, .ver etc",OptionMap::necessary);
  Option<deque<string> > par_file("par","list of parameter file",OptionMap::necessary);
  Option<string>         datafile("data","input data file",OptionMap::necessary);
  Option<string>         duplextype("duplextype","type of duplex, e.g. DNA or RNA",OptionMap::optional,"DNA");
  Option<double>         temp_err("ee","experimental temperature error in celsius",OptionMap::optional,0.0);
  Option<int>            random_seed("rs","random_seed",OptionMap::optional,1);
  Option<double>         salt("salt","salt concentration (only used for listing results)",OptionMap::optional);
  Option<double>         species_concentration("ct","fixed species concentration Ct (in 10^-6 M)",OptionMap::optional,0.0);
  Option<double>         multiply_sp_conc("mct","multiply species concentration Ct by a given factor",OptionMap::optional,1.0);
  Option<string>         calculation("calc","which cacluation to perform (minlocal,minglobal,prediction)",OptionMap::necessary,"prediction");
  Option<deque<string> > variable("var","list of variables to loop",OptionMap::optional);
  Option<string>         minimization_method("minmethod","select select minimization method (GSL or GBC)",OptionMap::optional,"GSL");
  Option<bool>           randomize_dataset("randomizedataset","randomize dataset",OptionMap::optional,false);
  Option<int>            max_random_rounds("mrr","maximum number of local minimization rounds (option calc=minglobal)",OptionMap::optional,10);
  Option<map<string,string> > dict("dict","Additions to nucleotide dictionary",OptionMap::optional);
  Option<double>         limit("limit","minimization limit",OptionMap::optional,0.01);
  Option<Range<double> > clength("cl","characteristic length",OptionMap::optional,10.0);
  Option<int>            max_eval("me","maximum number of evaluations",OptionMap::optional,std::numeric_limits<int>::max());
  Option<int>            max_min("mm","maximum number of minimizations",OptionMap::optional,2);
  Option<double>         av_diff_limit("adl","limit for total average difference",OptionMap::optional);
  Option<double>         av_stdev_limit("asl","limit for total standard deviation difference",OptionMap::optional);
  Option<bool>           generate_n_files("gn","generate files for each round of minimization",OptionMap::optional,true);

  Option<string>         par_id("parid","parameter id",OptionMap::optional,"vargibbs");
  Option<string>         debug("debug","select debuging",OptionMap::optional);
  Option<string>         sequence("seq","sequence main strand 5'->3'",OptionMap::optional);
  Option<string>         csequence("cseq","sequence complementary strand 3'->5'",OptionMap::optional);
  Option<double>         random_initial_values("randominitial","random initial values",0.0);
  //Unused option, reserved for future use
  Option<int>            seq_length("sl","fixed sequence length",OptionMap::optional,0);

  
  OptionMap::arguments(argc,argv);
  OptionMap::scan_arguments();
  if ((calculation.Value == "minlocal") || (calculation.Value == "minglobal"))
    {
    variable.turn_necessary(true);
    if (calculation.Value == "minglobal")
      {
      temp_err.turn_necessary(randomize_dataset.Value==1);
      random_initial_values.turn_necessary(!randomize_dataset.Provided);
      randomize_dataset.turn_necessary(!random_initial_values.Provided);
      }
    }
  if (calculation.Value == "prediction")
    {
    species_concentration.turn_necessary(sequence.Provided);
    datafile.turn_optional(sequence.Provided);
    sequence.turn_necessary(!datafile.Provided);
    }
  
  OptionMap::check_if_all_given();
  OptionMap::show_options();

  if (dict.provided())
    {
    map<string,string>::const_iterator mit;
    for(mit=dict.Value.begin(); mit!= dict.Value.end(); ++mit)
      {
      Nucleotide<>::dictionary.add(mit->first[0],mit->first[0]);
      Nucleotide<>::dictionary.add(mit->second[0],mit->second[0]);
      Nucleotide<>::dictionary.complementary_pair(mit->second[0],mit->first[0]);
      }
    }

  if (duplextype.Value=="RNA")
    {
    Nucleotide<>::dictionary.add('U','U');
    Nucleotide<>::dictionary.rm('T');
    Nucleotide<>::dictionary.complementary_pair('A','U');
    std::cout << Nucleotide<>::dictionary;
    }


  min_function_type pregress;
  if (minimization_method.provided()) pregress.minimization_method(minimization_method.Value);
  pregress.Avoid_negative_parameters=false;

  pregress.MaxEvaluations=max_eval;

  pregress.MaxMinimizations=max_min;

  if (max_random_rounds.provided()) pregress.MaxRandomMinimizations=max_random_rounds.Value;
  
  if (av_diff_limit.provided()) pregress.Limit_average_difference=av_diff_limit.Value;
  if (av_stdev_limit.provided()) pregress.Limit_standard_deviation_difference=av_stdev_limit.Value;
  
  
  if (debug.provided()) 
    {
    if (debug.Value == std::string("ParameterSearch.Evaluation")) pregress.Evaluation.Calculation.Debug=true;
    if (debug.Value == std::string("ParameterSearch")) pregress.Debug=true;
    }

  deque<string>::iterator pit;
  for (pit=par_file.Value.begin(); pit != par_file.Value.end(); ++pit)
    {
    pregress.Evaluation.Calculation.get_parameters(*pit);
    }
    
  if(calculation.Value=="prediction")
    {
    IndividualNearestNeighbours<> gfe;
    FreeEnergyModel<IndividualNearestNeighbours<> > gibbs;
    gibbs.model=&gfe;
    ofstream outfile; outfile.open(((string)output_name+".dat").c_str());
    ofstream verfile; verfile.open(((string)output_name+".ver").c_str());

    deque<string>::iterator it;
    for (it=par_file.Value.begin(); it != par_file.Value.end(); ++it) gibbs.model->get_parameters(*it);

    if(datafile.provided())
      {
      gibbs.dataset.read_xml((string)datafile,0,outfile,species_concentration.Value,multiply_sp_conc.Value);
      }
    else
      {
      cout << "Calculating for single sequence " << sequence.Value << "/" << csequence.Value << std::endl;
      gibbs.dataset.add_temperature(sequence.Value,csequence.Value,salt.Value,0.0,species_concentration.Value);
      }
            
    gibbs.predict_all();
    gibbs.dataset.print_temperatures(outfile);
    gibbs.dataset.verify(verfile);
    }

  if(calculation.Value=="minsimple" || calculation.Value=="minlocal")
    {
    if(species_concentration.provided()) pregress.Evaluation.Regression.dataset.Species_concentration=species_concentration;
    if(datafile.provided()) pregress.Evaluation.read_xml(datafile,seq_length.Value);
    pregress.basefile(output_name);
    pregress.assign(variable.Value);
    pregress.Generate_files_for_each_round=generate_n_files.Value;
    pregress.minimization(par_id,salt,clength.Value,limit.Value);
    }

  if(calculation.Value=="minmult" || calculation.Value=="minglobal")
    {
    if(species_concentration.provided()) pregress.Evaluation.Regression.dataset.Species_concentration=species_concentration;
    if(datafile.provided()) pregress.Evaluation.read_xml(datafile,seq_length.Value);
    pregress.basefile(output_name);
    pregress.assign(variable.Value);
    pregress.Generate_files_for_each_round=generate_n_files.Value;
    if (randomize_dataset.Value)
      {
      pregress.Randomise_dataset=randomize_dataset.Value;
      pregress.Evaluation.Regression.dataset.Temperature.error=temp_err;
      }
    if (random_initial_values.provided())
      {
      pregress.Randomise_initial_parameters=true;
      pregress.Initial_parameter_variation=random_initial_values.Value;
      }


    pregress.minimization_random(random_seed,par_id,salt,clength.Value,limit.Value);
    }

  std::cout << "Log: " << JobControl::LogFileName << " ";
  JobControl::summary(std::cout) << std::endl;
  JobControl::summary();
  JobControl::status("running",false);

  return 0;
}

